;; Generated package description from leaf.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "leaf" "4.5.5.0.20260508.153103" "Simplify your init.el configuration, extended use-package" '((emacs "24.1")) :commit "1e363fe114310a6af5f3f770be795657307ad3d1" :authors '(("Naoya Yamashita" . "conao3@gmail.com")) :maintainer '("Naoya Yamashita" . "conao3@gmail.com") :keywords '("lisp" "settings") :url "https://github.com/conao3/leaf.el")
